use crate::lib::entity::{Ant, Egg, Queen};
use crate::lib::point::{astar, Point};
use crate::lib::screen::{BoardCommand, Screen};
use crate::lib::world::World;
use rand::prelude::SliceRandom;
use rand::thread_rng;

pub trait AI {
    fn step(&mut self, b: &Screen, w: &mut World) -> BoardCommand;
    fn plan(&mut self, w: &World) {}
}

#[derive(Clone, Debug)]
pub enum AIGoal {
    Reach(Point),
    //Pickup(Point),
    //Drop(),
    Idle,
}

impl AI for Ant {
    fn plan(&mut self, w: &World) {
        // check last part of plan
        if let Some(goal) = self.plan.last() {
            match goal {
                AIGoal::Reach(target) => {
                    if self.pos == *target {
                        self.plan.pop();
                    }
                }
                AIGoal::Idle => {}
            }
        }
    }

    // return the next move for this ant
    fn step(&mut self, b: &Screen, w: &mut World) -> BoardCommand {
        let goal = match self.plan.last() {
            Some(g) => g,
            None => &AIGoal::Idle,
        };

        let choice = match goal {
            AIGoal::Idle => {
                // valid_movements does not return any diggables!
                let valid = w.get_valid_movements(&self.pos, b, true);
                let mut rng = thread_rng();
                valid.choose(&mut rng).cloned()
            }
            AIGoal::Reach(target) => {
                // here astar only produces a path between the target & pos without digging
                let result = astar(&self.pos, &target, Some(w), Some(b));
                let mut movements = match result {
                    Ok(m) => m,
                    Err(_) => vec![]
                };
                movements.pop()
            }
        };

        if !choice.is_none() {
            let pos = choice.unwrap();
            if w.cleared.contains(&pos) {
                self.pos = pos;
            } else {
                if w.is_safe_to_dig(&pos, &b) {
                    w.clear(pos);
                }
            }
        }

        BoardCommand::Noop
    }
}

impl AI for Egg {
    fn step(&mut self, b: &Screen, w: &mut World) -> BoardCommand {
        if self.counter > 0 {
            self.counter -= 1;
            return BoardCommand::Noop;
        } else {
            BoardCommand::Hatch(self.id, self.queen_id)
        }
    }
}

impl AI for Queen {
    fn step(&mut self, b: &Screen, w: &mut World) -> BoardCommand {
        let valid: Vec<Point> = w.get_valid_movements(&self.pos, b, false);
        let mut rng = thread_rng();

        let choice = valid.choose(&mut rng).cloned();

        if !choice.is_none() {
            let pos = choice.unwrap();
            if w.cleared.contains(&pos) {
                // choose between laying an egg and moving
                if self.egg_count < 3 {
                    self.egg_count += 1;
                    return BoardCommand::LayEgg(pos, self.id);
                } else {
                    self.pos = pos;
                }
            }
        }
        BoardCommand::Noop
    }
}
