use crate::lib::screen::Screen;
use crate::lib::world::World;
use std::collections::{HashMap, HashSet};

#[derive(Hash, PartialEq, Eq, Clone, Debug, Copy)]
pub struct Point(pub i32, pub i32);

impl Point {
    pub fn get_neighbors(&self) -> Vec<Point> {
        vec![self.left(), self.right(), self.up(), self.down()]
    }

    pub fn left(&self) -> Point {
        Point(self.0 - 1, self.1)
    }

    pub fn is_below(&self, other: Point) -> bool {
        other.down() == *self
    }

    pub fn right(&self) -> Point {
        Point(self.0 + 1, self.1)
    }

    pub fn uldiag(&self) -> Point {
        Point(self.0 - 1, self.1 - 1)
    }

    pub fn bldiag(&self) -> Point {
        Point(self.0 + 1, self.1 - 1)
    }

    pub fn brdiag(&self) -> Point {
        Point(self.0 + 1, self.1 + 1)
    }

    pub fn urdiag(&self) -> Point {
        Point(self.0 - 1, self.1 + 1)
    }
    // y values are reversed in ncurses
    pub fn down(&self) -> Point {
        Point(self.0, self.1 + 1)
    }

    pub fn up(&self) -> Point {
        Point(self.0, self.1 - 1)
    }
}

fn manhattan(c1: &Point, c2: &Point) -> i32 {
    return (c2.0 - c1.0).abs() + (c2.1 - c1.1).abs();
}

fn weigh_point(w: &World, p: &Point) -> i32 {
    if w.is_cleared(p) {
        3
    } else {
        15
    }
}

pub struct PathNotFoundError;

pub fn astar(start: &Point, goal: &Point, w: Option<&World>, s: Option<&Screen>) -> Result<Vec<Point>, PathNotFoundError> {
    let mut open_set: HashSet<Point> = HashSet::new();
    open_set.insert(*start);

    let mut came_from: HashMap<Point, Point> = HashMap::new();

    let mut gscore: HashMap<Point, i32> = HashMap::new();
    gscore.insert(*start, 0);

    let mut fscore: HashMap<Point, i32> = HashMap::new();
    fscore.insert(*start, manhattan(&start, &goal));

    let mut answer = Vec::new();

    let mut current: Point = *start;
    while !open_set.is_empty() {
        let mut min_score = i32::MAX;
        for c in open_set.iter() {
            let val = fscore[c];
            current = if val < min_score { c.clone() } else { current };
            min_score = if val < min_score { val } else { min_score };
        }

        if current == *goal {
            answer.push(current.clone());
            while came_from.contains_key(&current) {
                current = came_from[&current].clone();
                if current != *start {
                    answer.push(current.clone());
                }
            }
            return Ok(answer);
        }

        open_set.remove(&current);

        let current_gscore = gscore[&current];

        let all_neighbors: Vec<Point> = match w {
            Some(w) => current
                .get_neighbors()
                .iter()
                .filter(|p| w.is_valid_movement(&current, p, s.unwrap()))
                .map(|e| e.clone())
                .collect(),
            None => current.get_neighbors(),
        };

        for neighbor in all_neighbors.iter() {
            let neighbor_score = if gscore.contains_key(&neighbor) {
                gscore[&neighbor]
            } else {
                i32::MAX
            };

            let weight = match w {
                Some(w) => weigh_point(&w, neighbor),
                None => 1,
            };

            let score = current_gscore + weight;

            if score < neighbor_score {
                gscore.insert(neighbor.clone(), score);
                fscore.insert(neighbor.clone(), score + manhattan(&neighbor, &goal));
                came_from.insert(neighbor.clone(), current.clone());
                open_set.insert(neighbor.clone());
            }
        }
    }

    Err(PathNotFoundError)
}

#[test]
fn test_astar() {
    let start = Point(0, 0);

    let goal = Point(10, 10);
    let answers = astar(&start, &goal, None, None);
    
    match answers {
        Ok(ans) => {
            for a in ans.iter() {
                println!("{:?}", &a);
            }     
        },
        Err(_) => {
            panic!("Path not found. That shouldn't happen!")
        }
    }

    
}
